package net.sergeych.bipack

import kotlinx.serialization.SerializationStrategy
import kotlinx.serialization.descriptors.SerialDescriptor
import kotlinx.serialization.encoding.AbstractEncoder
import kotlinx.serialization.encoding.CompositeEncoder
import kotlinx.serialization.modules.EmptySerializersModule
import kotlinx.serialization.modules.SerializersModule
import kotlinx.serialization.serializer
import net.sergeych.bintools.*

class BipackEncoder(val output: DataSink) : AbstractEncoder() {

    private var nextIsUnsigned = false

    override fun encodeElement(descriptor: SerialDescriptor, index: Int): Boolean {
        return super.encodeElement(descriptor, index).also {
            nextIsUnsigned = descriptor.getElementAnnotations(index).any { it is Unsigned }
        }
    }

    override val serializersModule: SerializersModule = EmptySerializersModule
    override fun encodeBoolean(value: Boolean) = output.writeByte(if (value) 1 else 0)
    override fun encodeByte(value: Byte) = output.writeByte(value.toInt())
    override fun encodeShort(value: Short) = if (nextIsUnsigned)
        output.writeNumber(value.toUInt())
    else
        output.writeNumber(value.toInt())

    override fun encodeInt(value: Int) {
        if (nextIsUnsigned)
            output.writeNumber(value.toUInt())
        else
            output.writeNumber(value)
    }

    fun encodeUInt(value: UInt) = output.writeNumber(value)
    override fun encodeLong(value: Long) = if (nextIsUnsigned)
        output.writeNumber(value.toULong())
    else
        output.writeNumber(value)

    override fun encodeFloat(value: Float) = output.writeFloat(value)
    override fun encodeDouble(value: Double) = output.writeDouble(value)
    override fun encodeChar(value: Char) = output.writeNumber(value.code.toUInt())
    override fun encodeString(value: String) {
//        output.writeUTF(value)
        writeBytes(value.encodeToByteArray())
    }

    fun writeBytes(value: ByteArray) {
        output.writeNumber(value.size.toUInt())
        output.writeBytes(value)
    }

    override fun encodeEnum(enumDescriptor: SerialDescriptor, index: Int) = output.writeNumber(index.toUInt())

    override fun beginCollection(descriptor: SerialDescriptor, collectionSize: Int): CompositeEncoder {
        encodeUInt(collectionSize.toUInt())
        return this
    }

    override fun beginStructure(descriptor: SerialDescriptor): CompositeEncoder {
        // frame protection should start before anything else:
        val sink = if (descriptor.annotations.any { it is CrcProtected })
            CRC32Sink(output)
        else
            output
        // now it is safe to process anything else using `sink`. not the output!
        for (a in descriptor.annotations) {
            if (a is Framed) {
                sink.writeU32(
                    CRC.crc32(descriptor.serialName.encodeToByteArray())
                )
            } else if (a is Extendable) {
                sink.writeSmartUInt(descriptor.elementsCount.toUInt())
            }
        }
        return BipackEncoder(sink)
    }

    override fun endStructure(descriptor: SerialDescriptor) {
        if (output is CRC32Sink && descriptor.annotations.any { it is CrcProtected }) {
            output.writeU32(output.crc)
        }
        super.endStructure(descriptor)
    }

    override fun encodeNull() = encodeBoolean(false)
    override fun encodeNotNullMark() = encodeBoolean(true)

    companion object {
        fun <T> encode(serializer: SerializationStrategy<T>, value: T, sink: DataSink) {
            val encoder = BipackEncoder(sink)
            encoder.encodeSerializableValue(serializer, value)
        }

        fun <T> encode(serializer: SerializationStrategy<T>, value: T): ByteArray =
            ArrayDataSink().also { encode(serializer, value, it) }.toByteArray()

        inline fun <reified T> encode(value: T) = encode(serializer(), value)
        inline fun <reified T> encode(value: T, sink: DataSink) = encode(serializer(), value, sink)

    }
}
